part of dartscale;

/**
 * Represents an isolated Context in which Modules can be registered/unregistered
 * and started/stopped.
 * 
 * An Dart application could use several [Core] instances.
 */
class Core {
  
  final Map<String, Function> _registeredModules = new Map<String, Function>();
  final Map<String, dynamic> _runningModules = new Map<String, dynamic>();
  
  final Mediator mediator = new Mediator();
  
  /**
   * Register a new module in the [Core] by passing an constructor 
   * function as first parameter and the module name as second.
   * 
   * The constructor function will be used each time the module is started
   * and a new instance of it is needed, so this function should return a blank instance.
   * 
   * __no subclassing required__
   * The module class doesn't have to subclass any specific framework type,
   * but it has to have a method named _init_ which receives a [Sandbox] object
   * as single parameter and is invoked when the module is started by the [Core].
   * 
   * Also two methods are required: a _stop_ method without any parameters  
   * a _start_ method which will receive the options passed to the [Core]s [start], 
   * [startMany] or [startAll] method.  
   * 
   *     class MyModule {
   *       init(Sandbox sandbox) {
   *         // you may store the sandbox object in the instance here
   *       }
   *         
   *       start(options) {
   *         // do something
   *       }
   *         
   *       stop() {
   *          // remove HtmlElements, release ressources...
   *       }
   *     }
   *     
   *     core.register(() => new MyModule(), 'MyModule');
   *
   */
  register(Function constructor, String moduleName) {
    if (_registeredModules.containsKey(moduleName)) {
      throw new StateError("Module ${moduleName} already registered!");
    }
    
    _registeredModules[moduleName] = constructor;
  }
  
  /**
   * Remove the registered module from the [Core].
   */
  unregister(String moduleName) {
    if (!_registeredModules.containsKey(moduleName)) {
      throw new StateError("Module ${moduleName} not registered!");
    }
    
    _registeredModules.remove(moduleName);
  }
  
  /**
   * Start a module. By default the modules name is used as its Id.
   * You can provide a custom Id by passing as string as optional named argument.
   * 
   *     core.start('MyModule', options: {
   *       'container': querySelector('#myModule')
   *     }, id: 'myDifferentId');
   *
   */
  start(String moduleName, {String id: null, options: null}) {
    if (moduleName is String) {
      _start(moduleName, id, options);
    }
    else if (moduleName == null) {
			this.startAll(options);

    }
    else {
      throw new ArgumentError("Invalid Type given for argument moduleName");
    }
  }
  
  /**
   * Start all modules whose name are in the List of module names passed as first argument.
   * 
   * This will use the module names as their Ids by default.
   * 
   *     core.startMany(['Navbar', 'TodoList'], options: {
   *       'alertEl': querySelector('#alert')
   *     });
   *
   */
  startMany(List<String> moduleNames, [options]) {
		moduleNames.forEach((name) => _start(name, null, options));
  }
  
  /**
   * Start all modules registered in this [Core].
   * 
   * This will use the module names as their Ids by default.
   *  
   *     core.startAll({
   *       'alertEl': querySelector('#alert')
   *     });
   *
   */
  startAll([options]) {
    for (String name in _registeredModules.keys) {
      _start(name, null, options);
    }
  }
  
  _start(String moduleName, String id, options) {
    if (!_registeredModules.containsKey(moduleName)) {
      throw new StateError("Module ${moduleName} not registered!");
    }
    
    final String moduleId = id != null ? id : moduleName;
    
    if (_runningModules.containsKey(moduleId)) {
      throw new StateError("Module with id #${moduleId} already running!");
    }
    
		final dynamic moduleInstance = _registeredModules[moduleName]();
		moduleInstance.init(new Sandbox(this.mediator));
		moduleInstance.start(options);
    
    _runningModules[moduleId] = moduleInstance;
  }
  
  /**
   * Stop the currently running module which Id is passed as argument.
   *   
   *     core.stop('MyModule');
   * 
   */
  stop(String moduleId) {
    _stop(moduleId);
  }
  
  /**
   * Stop the modules which Id are in the [List] of Ids passed as argument.
   *    
   *     core.stopMany(['MyModule', 'MyDifferentId']);
   *
   */
  stopMany(List<String> moduleIds) {
		moduleIds.forEach((moduleId) => _stop(moduleId));
  }
  
  /**
   * Stop all currently running modules.
   *    
   *     core.stopAll();
   *
   */
  stopAll() {
    List<String> ids = [];
		
		// _stop manipulates the _runningModules HashMap,
    // thus it'd result in an error calling _stop in an iteration
    ids.addAll(_runningModules.keys);
		
    ids.forEach((moduleId) => _stop(moduleId));
  }
  
  _stop(String moduleId) {
    if (!_runningModules.containsKey(moduleId)) {
      throw new StateError("Module with id #${moduleId} not running!");
    }
    
    _runningModules.remove(moduleId).stop();
  }
  
  /**
   * Check if a module is registered.
   * 
   * If no argument is provided, a [List] of strings with the names of
   * all registered modules is returned.
   *     
   *     if (core.registered('MyModule')) {
   *       // do something
   *     }
   *     
   *     // or how about
   *     
   *     core.mediator('application', 'currentlyRegistered').add(core.registered());
   *
   */
  registered([String moduleName]) {
    if (moduleName != null) {
      return _registeredModules.containsKey(moduleName);
    }
    else {
      return _registeredModules.keys;
    }
  }
  
  /**
   * Check if a module is running.
   * 
   * If no argument is provided, a [List] of strings with the names of
   * all running modules is returned.
   *     
   *     if (core.running('MyModule')) {
   *       // do something
   *     }
   *     
   *     // or how about
   *     
   *     core.mediator('application', 'currentlyRunning').add(core.running());
   *
   */
  running([String moduleId = null]) {
    if (moduleId != null) {
      return _runningModules.containsKey(moduleId);
    }
    else {
      return _runningModules.keys;
    }
  }
}