part of dartscale;

/**
 * An object through which the module has access to several functions
 * of the Framework, such as the [Core]s [Mediator].
 * 
 * Also a [Sandbox] can have several plugins registered to have
 * fine grained controll to what the module has access.
 */
class Sandbox {
  
  final Map<Symbol, dynamic> _plugins = new Map<Symbol, dynamic>();
  
  Mediator _mediator;
	
	Mediator get mediator => _mediator;
  
  Sandbox(Mediator this._mediator);
	
  /**
   * __not used by your code!__
   */
  noSuchMethod(Invocation invocation) {
      var plugin = _plugins[invocation.memberName];
      
      if (invocation.isSetter) {
        throw new UnsupportedError("Setting Plugins is not allowed!");
      }
      
      if (plugin == null) {
        throw new NoSuchPluginError(invocation.memberName.toString());
      }
      
      if (invocation.isGetter || invocation.isAccessor) {
        return plugin;
      }
      else if (invocation.isMethod) {
        return Function.apply(plugin, invocation.positionalArguments, invocation.namedArguments);
      }
  }
  
  /**
   * Add a plugin which can be any object or function.
   * Later on it can be accessed like any property of the sandbox.
   * 
   *     var sandbox = new Sandbox(new Mediator());
   * 
   *     sandbox.addPlugin("testMethod", (arg1, arg2) {
   *       return "${arg1}+${arg2}";
   *     });
   * 
   *     var result = sandbox.testMethod("Hello", "World");
   *
   */
  addPlugin (String name, dynamic plugin) {
    _plugins[new Symbol(name)] = plugin;
  }
	
	/**
	 * Add a number of plugins, provided as [Map].
	 * This'll use [addPlugin] under the hood.
	 * 
	 *     var sandbox = new Sandbox(new Mediator());
   * 
   *     sandbox.addPlugins({
   *       "testMethod": (arg1, arg2) {
   *         return "${arg1} ${arg2}";
   *       },
   *       
   *       "add": (x, y) {
   *         return x + y;
   *       }
   *     });
   * 
   *     var result = sandbox.testMethod("Hello", "World");
	 * 
	 */
	addPlugins (Map<String, dynamic> plugins) {
		plugins.forEach((String name, dynamic plugin) => addPlugin(name, plugin));
	}
	
	/**
	 * Inherit all plugins from an existing [Sandbox] instance.
	 */
	inheritPlugins (Sandbox source) {
		_plugins.addAll(source._plugins);
	}
	
	/**
	 * Inherit the [Mediator]] from an existing [Sandbox] instance.
	 * 
	 * __Attention__  
	 * This will give modules which're using either of both sandboxes
	 * access to the same mediator (messagebus) and can lead to unexpected
	 * behaviour / bugs if you're not carefull.
	 */
	inheritMediator (Sandbox source) {
		_mediator = source._mediator;
	}
}